@@ -41,7 +41,11 @@ class Style(BaseModel):
     cmap: Any = 'inferno'
     sequential_cmap: Any = 'Greens'
     log_formatter: str = "\n@@ {time:x} AT: {time} | LEVEL: {level} | IN: {name}.{function}\n\n{message} |\n"
-
+    timeframe_prefix: str = 'from '
+    timeframe_between: str = ' to '
+    timeframe_suffix: str = ''
+    auto_locf_params: Dict[str, Any] = {'window': 3, 'impute_method': 'median',
+                                        'impute_direction': 'forward', 'add_noise': False}
 
     def translate(self, key: str, **kwargs) -> str:
         """ Helper function that allows Style objects to translate text according to the provided Rosetta mappings
@@ -80,4 +84,4 @@ class Style(BaseModel):
         else:
             raise ValueError(f"Unsure how to interpret 3+ inputs to override")
 
-        return Style(**{**self.dict(), **override_dict})
+        return self.__class__(**{**self.dict(), **override_dict})
