@@ -17,6 +17,7 @@ from pydantic import BaseModel
 from sklearn.feature_selection import SelectKBest, chi2
 from tqdm.auto import tqdm
 import matrixprofile
+import math
 
 
 class SVD(BaseModel):
@@ -360,16 +361,52 @@ class VectorMultiset(PickleUtils, Style, Rosetta):
             return to_return
 
     def matrix_profile_univariate(self, col_names: Union[str, Iterable[str]],
+                                  auto_locf: bool = False, suppress_nan_warning: bool = False,
                                   **kwargs) -> List[Tuple[Any, List[plt.Figure]]]:
         """
-        Very thin wrapper around matrixprofile
+        Very thin wrapper around matrixprofile with some warnings and opt-in preprocessing automation
 
         :param col_names: a column name (or list of column names) to be profiled
+        :param auto_locf:
         :param kwargs: keyword arguments passed through to matrixprofile.analyze()
         :return: List of outputs from matrixprofile.analyze()
         """
         queue: List[str] = [col_names] if isinstance(col_names, str) else col_names
-        return [matrixprofile.analyze(self.data.loc[:, q].tolist(), **kwargs) for q in queue]
+        results: List[Tuple[Any, List[plt.Figure]]] = []
+        try:
+            for q in queue:
+                ts: List[float] = self.data.loc[:, q].astype(float).tolist()
+
+                if not kwargs.get('preprocessing_kwargs') and any(np.isnan(x) for x in ts):
+                    if hasattr(self, 'auto_locf_params'):
+                        auto_locf_params: Dict[str, Any] = self.auto_locf_params
+                    else:
+                        auto_locf_params: Dict[str, Any] = {'window': 3, 'impute_method': 'median',
+                                                            'impute_direction': 'forward', 'add_noise': False}
+                    if auto_locf:
+                        kwargs.setdefault('preprocessing_kwargs', auto_locf_params)
+                    else:
+                        if not suppress_nan_warning:
+                            print("""
+    Warning: matrix profiling data with NaNs and no preprocessing specified
+             (potential anomalies or errors ahead)\n
+    > To suppress this warning in the future, specify `suppress_nan_warning=True`.
+    > One way to address NaNs is specifying a 'preprocessing_kwargs' dictionary, like:
+        {'window': 3, 'impute_method': 'median', 'impute_direction': 'forward', 'add_noise': False}
+    > Alternatively, you can specify `auto_locf=True` to use the config presets:\n""" +
+                                  "".join([f"       ~ {k}={v}\n" for k, v in auto_locf_params.items()]))
+                results.append(matrixprofile.analyze(ts, **kwargs))
+        except TypeError as e:
+            s: str = f"""
+            Encountered a TypeError in matrixprofile.analyze(). Hint: this might happen if you are
+            trying to use a custom parameter like 'use_right_edge' with the standard matrixprofile
+            package. If that is the case, try running:\n
+              pip uninstall matrixprofile && pip install git+https://github.com/mitchellpkt/matrixprofile
+            \nOriginal error: {e}"""
+            print(s)
+            raise TypeError(s)
+
+        return results
 
     def calculate_stumpy_profile_univariate(self, col_name: str, window_size: int, **kwargs) -> np.ndarray:
         """
@@ -461,7 +498,8 @@ class SlidingWindowResults(VectorMultiset):
         return visualize_x_y(x_arrays, y_arrays, legend_strings=labels, style=Style(**self.dict()), **kwargs)
 
     def plot_pdfs(self, col_name: str, legend_override: List[str] = None, **kwargs) -> plt.Figure:
-        """ Plot the probability density function(s) of the sliding window results. Useful kwargs: [cumulative, density]
+        """
+        Plot the probability density function(s) of the sliding window results. Useful kwargs: [cumulative, density]
 
         :param col_name: which field should be plotted
         :param legend_override: custom legend labels if desired (must have same length as number of unique widths)
@@ -481,6 +519,13 @@ class SlidingWindowResults(VectorMultiset):
         return visualize_1d_distribution(data_sets, legend_strings=labels, **kwargs)
 
     def heatmap_feature(self, col_name: str, **kwargs) -> plt.Figure:
+        """
+        Returns a heatmap of some feature as a function of window start time and start width
+
+        :param col_name: which column to plot
+        :param kwargs: additional keyword arguments for surface method
+        :return: figure handle of the plot
+        """
         return self.surface(self.window_start_col_name, self.window_width_col_name, col_name, **kwargs)
 
 
@@ -511,6 +556,12 @@ class VectorSequence(VectorMultiset):
 
     def __init__(self, skip_vector_sequence_init: bool = False, **kwargs: Any):
         super().__init__(**kwargs)
+
+        # Cast to numeric if possible
+        if not kwargs.get('disable_auto_conversion_to_numeric', False) and (self.data is not None):
+            self.cast_to_numeric()
+
+        # Handle the other processing specific to ordered data
         if (self.data is not None) and len(self.data) and (not skip_vector_sequence_init):
             if (not self.basis_col_name) or (self.basis_col_name not in self.data.keys()):
                 raise ValueError(f"{self.basis_col_name=} not in known keys: {self.data.keys().tolist()=}\n" +
@@ -558,7 +609,7 @@ class VectorSequence(VectorMultiset):
         return fill_ratio(self.data.loc[:, self.basis_col_name])
 
     def slice(self, start_at: Any = None, stop_at: Any = None, inplace: bool = False, reset_index: bool = True,
-              return_type: str = None) -> Union[None, pd.DataFrame, Tuple[Iterable, Iterable], Any]:
+              return_type: str = None, **kwargs) -> Union[None, pd.DataFrame, Tuple[Iterable, Iterable], Any]:
         """ Slices the VectorSequence according to the basis
 
         # TODO have slice return self type for convenience
@@ -568,6 +619,7 @@ class VectorSequence(VectorMultiset):
         :param inplace: slice inplace or return the result (default --> return)
         :param return_type: how to return the data ('VectorMultiset', 'VectorSequence', 'dataframe', etc)
         :param reset_index: indicate whether or not to reset the data frame index (default --> True)
+        :param kwargs: keyword arguments passed through on init of non-dataframe objects
         :return: VectorSequence if not inplace
         """
         df: pd.DataFrame = deepcopy(self.data)
@@ -581,18 +633,21 @@ class VectorSequence(VectorMultiset):
             in_range.reset_index(drop=True, inplace=True)
         if inplace:
             self.data = in_range
-
-        if not return_type:
-            return self.__class__(basis_col_name=self.basis_col_name, name_root=self.name_root, data=in_range)
         else:
-            if 'sequence' in (return_type_lower := return_type.lower()):
-                return VectorSequence(basis_col_name=self.basis_col_name, name_root=self.name_root, data=in_range)
-            elif 'dataframe' in return_type_lower:
-                return in_range
-            elif 'timeseries' in return_type_lower:
-                return Timeseries(basis_col_name=self.basis_col_name, name_root=self.name_root, data=in_range)
+            if not return_type:
+                return self.__class__(basis_col_name=self.basis_col_name, name_root=self.name_root,
+                                      data=in_range, **kwargs)
             else:
-                raise ValueError(f"Unknown return type: {return_type}")
+                if 'sequence' in (return_type_lower := return_type.lower()):
+                    return VectorSequence(basis_col_name=self.basis_col_name, name_root=self.name_root,
+                                          data=in_range, **kwargs)
+                elif 'dataframe' in return_type_lower:
+                    return in_range
+                elif 'timeseries' in return_type_lower:
+                    return Timeseries(basis_col_name=self.basis_col_name, name_root=self.name_root,
+                                      data=in_range, **kwargs)
+                else:
+                    raise ValueError(f"Unknown return type: {return_type}")
 
     def index_to_basis(self) -> None:
         """  Adds (or overwrites) 'basis' column with a default index [0, 1, 2, ...] """
@@ -665,7 +720,18 @@ class VectorSequence(VectorMultiset):
         """
         window_data: VectorSequence = self.slice(start_at=start_at, stop_at=start_at + window_width,
                                                  inplace=False, reset_index=True)
-        return {**function(window_data, *args, **kwargs), 'window_start': start_at, 'window_width': window_width}
+        try:
+            return {**function(window_data, *args, **kwargs), 'window_start': start_at, 'window_width': window_width}
+        except IndexError as e:
+            window_description: str = f"{start_at=}\n{window_width=}\n"
+            if not len(window_data):
+                raise IndexError("\nCaught IndexError.\n\nHint: there were no data points in this window:\n" +
+                                 f"{window_description}\n(The best way to resolve this is probably to tweak\n" +
+                                 f"your eval function to gracefully handle empty slices)\n\nOriginal error:\n{e}")
+            else:
+                raise IndexError(f"Eval function had IndexError at window:\n{window_description}\nOriginal error:\n{e}")
+        except TypeError as e:
+            raise TypeError(f"Caught type error (below). Hint: your eval function needs to return a dictionary.\n\n{e}")
 
     #                                       vv sequence vv       vv args vv          vv kwargs vv
     def sliding_window(self, function: Callable[[Any, Union[Tuple[Any], List[Any]], Dict[str, Any]], Dict[str, Any]],
@@ -916,11 +982,179 @@ class VectorSequence(VectorMultiset):
                 kwargs: Dict[str, Any] = {'exclude_cols': self.basis_col_name}
         return correlation_matrix(self.data, **kwargs)
 
+    def downsample_and_plot_matrix_profile_univariate(self, downsample_interval: float,
+                                                      col_names: Union[str, Iterable[str]], diff: bool = False,
+                                                      downsample_method: str = 'by_basis',
+                                                      **kwargs) -> List[Tuple[Any, List[plt.Figure]]]:
+        """
+        Downsamples and optionally diffs data before applying the matrix profile (to which kwargs are passed through)
+
+        :param downsample_interval: interval by which to downsample (either by N rows, or by some basis interval)
+        :param downsample_method: 'by_basis' or 'by_row'
+        :param col_names: a column name (or list of column names) to be profiled
+        :param diff: whether to take the diff of the rows (coming soon - fractional differentiation??)
+        :param kwargs: keyword arguments passed through to matrixprofile.analyze()
+        :return: List of outputs from matrixprofile.analyze()
+        """
+        downsampled = self.downsample(downsample_interval, method=downsample_method, inplace=False)  # self type
+        if diff:
+            downsampled.data = downsampled.data.diff()
+            downsampled.data.dropna(inplace=True)
+        downsampled.data.reset_index(inplace=True)
+        return downsampled.matrix_profile_univariate(col_names, **kwargs)
+
+    def plot_fluss_semantic_segmentation(self, col_name: str = None, fluss_width: int = 15,
+                                         n_regimes: int = 2, excl_factor: float = 1, **kwargs) -> plt.Figure:
+        """
+        Wrapper for STUMPY's FLUSS algonithm + plots
+
+        :param col_name: name of the column to plot
+        :param fluss_width: FLUSS window size
+        :param n_regimes: number of regimes in the data
+        :param excl_factor: factor to correct for the start and end of the arc curve
+        :param kwargs: additional keyword arguments for stumpy and plots (figsize, title, ...)
+        :return: figure handle showing the data and the analysis results
+        """
+        figsize: Any = kwargs.pop('figsize', self.figsize)  # typically a 2-element tuple or list
+        title: str = kwargs.pop('title', self.translate(self.name_root))
+
+        # If column name not specified and only one non-basis column, infer that we should use that one
+        if not col_name:
+            if len(self.data.keys()) == 2:
+                col_name: str = [x for x in self.data.keys() if x != self.basis_col_name][0]
+            else:
+                raise ValueError(f"For fluss, please specify `col_name=X` with X from: {list(self.data.keys())}")
+
+        # Calculate the matrix profile and FLUSS results
+        data_array: np.ndarray = self.data.loc[:, col_name].to_numpy()
+        mp: np.ndarray = stumpy.stump(data_array, m=fluss_width, **kwargs)
+        cac, regime_locations = stumpy.fluss(mp[:, 1], L=fluss_width, n_regimes=n_regimes, excl_factor=excl_factor)
+
+        # Plot the results
+        f, axs = plt.subplots(2, sharex=True, gridspec_kw={'hspace': 0}, figsize=figsize, facecolor=self.facecolor)
+        plt.suptitle(title, fontsize=self.title_fontsize)
+        axs[0].plot(range(data_array.shape[0]), data_array, color=self.color)
+        axs[0].axvline(x=regime_locations[0], linestyle="dashed")
+        axs[1].plot(range(cac.shape[0]), cac, color='k')
+        axs[1].axvline(x=regime_locations[0], linestyle="dashed")
+        return f
+
+    def human_time_start_and_stop(self, **kwargs) -> Tuple[str, str]:
+        """
+        Helper function that returns the min and max value of the basis column converted to human-readable string
+
+        :param kwargs: keyword arguments for human_time()
+        :return: tuple of strings like ('2025-03-21', '2028-07-04')
+        """
+        return self.human_time(min(self.data.loc[:, self.basis_col_name].astype(float).dropna().tolist()), **kwargs), \
+               self.human_time(max(self.data.loc[:, self.basis_col_name].astype(float).dropna().tolist()), **kwargs)
+
+    def human_timeframe(self, prefix: str = None, between: str = None, suffix: str = None, **kwargs) -> str:
+        """
+        Formatting helper function for ensuring consistency across timeframe descriptions, for example
+
+            ``From 2025-03-21 to 2028-07-04``
+
+        :param prefix: text to precede a time/date range
+        :param between: text to go between the times/dates
+        :param suffix: text to follow the time/date range
+        :param kwargs: additional keyword arguments for human_time, e.g. formatter = '%Y-%m-%d %H:%M:%S'
+        :return: a single string describing the timeframe
+        """
+        kwargs.setdefault('formatter', self.formatter)
+        if prefix is None:
+            prefix: str = self.timeframe_prefix
+        if between is None:
+            between: str = self.timeframe_between
+        if suffix is None:
+            suffix: str = self.timeframe_suffix
+        start_time_string, stop_time_string = self.human_time_start_and_stop(**kwargs)
+        return f"{prefix}{start_time_string}{between}{stop_time_string}{suffix}"
+
 
 class Timeseries(VectorSequence):
     """ Thin wrapper for VectorSequence in the context of time  """
     basis_col_name: str = 'timestamp'
 
+    def calc_weights(self, lowest_weight: float = 0.5, decay_timeframe_sec: float = None, show_plots: bool = False,
+                     method: str = 'linear', post_window_weight: float = None, base: float = np.e) -> List[float]:
+        """
+        Calculates a vector of weights from an exponential or linear decay
+
+        :param lowest_weight: what should be the minimum weight? (i.e. at the end of the decay timeframe)
+        :param decay_timeframe_sec: how long should the decay window be?
+        :param post_window_weight: what should weights be after the decay window? Set to 0 to forget anything older.
+        :param show_plots: whether to show plots of the weights
+        :param method: whether the decay should be linear or exponential
+        :param base: the base for the decay, but I don't think it matters
+        :return: a list of weights
+        """
+        min_timestamp: float = min(self.data.loc[:, self.basis_col_name])
+        max_timestamp: float = max(self.data.loc[:, self.basis_col_name])
+
+        # If decay_timeframe_sec not specified, apply to the whole data set
+        if decay_timeframe_sec is None:
+            decay_timeframe_sec = max_timestamp - min_timestamp
+
+        # If post_window_weight not specified, carry the last weight forward
+        if post_window_weight is None:
+            post_window_weight = lowest_weight
+
+        # Make sure that we don't collide on column name
+        age_col_name: str = 'age'
+        i: int = 0
+        while age_col_name in self.data.keys():
+            age_col_name += '_'
+            i += 1
+            if i > 10:  # this 10 is not a magic number, I just picked age_, age__, age___, up to 10 as excessive
+                raise NotImplementedError(f"Having {i} age columns with trailing underscores is not allowed...")
+
+        # Calculate age in a copy of the data frame
+        df: pd.DataFrame = deepcopy(self.data)
+        df[age_col_name] = max_timestamp - df.loc[:, self.basis_col_name]
+
+        # Calculate the weights (not taking into account the decay timeframe)
+        if 'lin' in method.lower():  # 'lin' or 'linear'
+            #  y = m * x + b
+            slope: float = (lowest_weight - 1) / decay_timeframe_sec
+            weights_all: List[float] = [slope * x + 1 for x in df.loc[:, age_col_name]]
+        elif 'exp' in method.lower():  # 'exp' or 'exponential'
+            #  y = b ^ (-k * x) where k is the decay constant and x is the age
+            decay_constant: float = -1 * math.log(lowest_weight, base) / decay_timeframe_sec
+            weights_all: List[float] = [base ** (-1 * decay_constant * tau) for tau in df.loc[:, age_col_name]]
+        else:
+            raise ValueError(f"Unknown method {method}, please try 'linear' or 'exponential'")
+
+        # Replace any out-of-timeframe data points with the post_window_weight
+        ages_and_weights: zip = zip(df.loc[:, age_col_name], weights_all)
+        weights_raw: List[float] = [w if a < decay_timeframe_sec else post_window_weight for a, w in ages_and_weights]
+        normalization_factor = sum(weights_raw)
+        final_weights: List[float] = [w / normalization_factor for w in weights_raw]
+
+        # Show plots of weights for the curious user
+        if show_plots:
+            self.scatter(df.loc[:, self.basis_col_name].tolist(), final_weights,
+                         xlabel=self.translate(self.basis_col_name),
+                         ylabel='weights', title='weights vs timestamps')
+
+            self.scatter(df.loc[:, age_col_name].tolist(), final_weights, xlabel='age',
+                         ylabel='weights', title='weights vs ages')
+            plt.show()
+
+        return final_weights
+
+    def weighted_mean(self, col_name: str, **kwargs) -> float:
+        """
+        Calculates the weighted mean of the provided column name. kwargs are passed to calc_weights
+
+        :param col_name: which column to calculate the weight for
+        :param kwargs: kwargs for calc_weights, e.g. lowest_weight, decay_timeframe_sec, post_window_weight, method
+        :return: weighted mean of that column
+        """
+
+        weights = self.calc_weights(**kwargs)
+        return sum([weight * val for weight, val in zip(weights, self.data.loc[:, col_name])])
+
 
 ################
 # Helpers
